#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <linux/if.h>
#include <linux/if_tun.h>

struct TunFrame
{
   uint16_t flags;
   uint16_t proto;
};
/* https://www.kernel.org/doc/Documentation/networking/tuntap.txt */
int tun_alloc(char *dev)
{
    struct ifreq ifr;
    int fd, err;

    printf("sizeof(TunFrame)=%ld\n",sizeof(struct TunFrame));
    if((fd = open("/dev/net/tun", O_RDWR)) < 0)
    {
       perror("open(/dev/net/tun)");
       return -1;
    }
    memset(&ifr,'\0', sizeof(ifr));
    /* Flags: IFF_TUN   - TUN device (no Ethernet headers)
     *        IFF_TAP   - TAP device
     *
     *        IFF_NO_PI - Do not provide packet information
     */
     ifr.ifr_flags = IFF_TUN;
     if( *dev )
     {
         strncpy(ifr.ifr_name, dev, IFNAMSIZ);
     }
     if( (err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0 )
     {
        perror("ioctl(TUNSETIFF)");
	close(fd);
	return err;
     }
     strcpy(dev,ifr.ifr_name);
     return fd;
}

char buffer[4096];

int main(int argc,char **argv)
{
   char devname[IFNAMSIZ];
   int fd; 
   ssize_t nr;

   if(argc < 2)
   {
      fprintf(stderr,"Aufruf: %s tun0\n",argv[0]);
      return -1;
   }
   strncpy(devname,argv[1],IFNAMSIZ);
   fd=tun_alloc(devname);
   if(fd < 0) 
   {
      fprintf(stderr,"Creating device %s failed\n",devname);
      return -1;
   }
   nr = read(fd,buffer,4096);
   while(nr >=0)
   {
      struct in_addr srcaddr;
      struct in_addr destaddr;
      char srcaddra[32];
      char destaddra[32];

      printf("%ld bytes read\n",nr);
      struct TunFrame *tunframe = (struct TunFrame *)buffer;
      printf("Flags=%u Proto=%u (flags=%u proto=%u)\n",ntohs(tunframe->flags),ntohs(tunframe->proto),tunframe->flags,tunframe->proto);
      struct iphdr *iphdr = (struct iphdr *)(buffer+sizeof(struct TunFrame));
      srcaddr.s_addr = iphdr->saddr;
      destaddr.s_addr = iphdr->daddr;
      strncpy(srcaddra,inet_ntoa(srcaddr),32);
      strncpy(destaddra,inet_ntoa(destaddr),32);
      printf("ihl=%u version=%u protocol=%u src=%s dest=%s\n",iphdr->ihl,iphdr->version,iphdr->protocol,
		      srcaddra,destaddra);
      nr = read(fd,buffer,4096);
   }
   if(nr < 0)
   {
      perror("read");
   }

   close(fd);
}
